"""
MCP Tool Adapter for Google Gemini.

This adapter converts remote MCP server tools to Gemini-compatible tool definitions
and processes Gemini's function calls to HTTP requests to the MCP server.

Requires the optional `google-genai` dependency:
    pip install mcphero[google-genai]
"""

from __future__ import annotations

try:
    from google.genai import types  # pyright: ignore[reportMissingImports]
except ImportError as e:
    from mcphero.exceptions import INSTALL_GOOGLE_GENAI

    raise ImportError(INSTALL_GOOGLE_GENAI) from e

import httpx

from mcphero.adapters.base_adapter import BaseAdapter


class MCPToolAdapterGemini(BaseAdapter):
    """
    Adapter that converts remote MCP server tools to Gemini-compatible tool definitions.
    Also converts Gemini's function calls to HTTP requests to the MCP server.

    Requires the optional `google-genai` dependency:
        pip install mcphero[google-genai]

    Usage with google-genai SDK:

        from google import genai
        from google.genai import types

        adapter = MCPToolAdapterGemini("https://api.mcphero.app/mcp/your-server-id")

        # Get tool definitions
        tool = await adapter.get_tool()

        # Create Gemini client and make request
        client = genai.Client(api_key="...")
        response = client.models.generate_content(
            model="gemini-2.5-flash",
            contents="What's the weather?",
            config=types.GenerateContentConfig(
                tools=[tool],
                automatic_function_calling=types.AutomaticFunctionCallingConfig(
                    disable=True
                ),
            ),
        )

        # Process function calls if present
        if response.function_calls:
            results = await adapter.process_function_calls(response.function_calls)
            # Add results to conversation and continue
    """

    async def get_function_declarations(self) -> list[types.FunctionDeclaration]:
        """
        Fetch tools from MCP server and convert them to Gemini FunctionDeclaration objects.

        Returns:
            List of FunctionDeclaration objects.
        """
        mcp_tools = await self.make_request("/tools")

        declarations: list[types.FunctionDeclaration] = []
        for tool in mcp_tools:
            declaration = types.FunctionDeclaration(
                name=tool["name"],
                description=tool.get("description", ""),
                parameters=tool.get(
                    "inputSchema",
                    {
                        "type": "object",
                        "properties": {},
                    },
                ),
            )
            declarations.append(declaration)

        return declarations

    async def get_tool(self) -> types.Tool:
        """
        Fetch tools from MCP server and return as a Gemini Tool object.

        This returns a Tool that can be passed directly to GenerateContentConfig.

        Returns:
            A Tool object containing all function declarations.
        """
        declarations = await self.get_function_declarations()
        return types.Tool(function_declarations=declarations)

    async def process_function_calls(
        self,
        function_calls: list[types.FunctionCall],
        return_errors: bool = True,
    ) -> list[types.Content]:
        """
        Process Gemini's function calls to invoke the tools via HTTP.

        Args:
            function_calls: List of FunctionCall objects from the Gemini response.
            return_errors: If True, include error responses for failed calls.

        Returns:
            List of Content objects containing function responses that can be
            added to the conversation.

        Example:
            if response.function_calls:
                results = await adapter.process_function_calls(response.function_calls)

                # Add to conversation:
                contents.append(response.candidates[0].content)  # Model's function call
                contents.extend(results)  # Function responses
        """
        results: list[types.Content] = []

        for fc in function_calls:
            tool_name = fc.name
            arguments = fc.args if fc.args else {}
            call_id = getattr(fc, "id", None)

            try:
                result = await self.make_request(
                    f"/tools/{tool_name}/call",
                    method="POST",
                    data={"arguments": arguments},
                )

                function_response = types.FunctionResponse(
                    name=tool_name,
                    response={"result": result}
                    if not isinstance(result, dict)
                    else result,
                    id=call_id,
                )

                content = types.Content(
                    role="user",
                    parts=[
                        types.Part.from_function_response(
                            name=function_response.name,
                            response=function_response.response,
                        )
                    ],
                )
                results.append(content)

            except httpx.HTTPError as e:
                if return_errors:
                    function_response = types.FunctionResponse(
                        name=tool_name,
                        response={"error": f"HTTP error calling tool: {str(e)}"},
                        id=call_id,
                    )
                    content = types.Content(
                        role="user",
                        parts=[
                            types.Part.from_function_response(
                                name=function_response.name,
                                response=function_response.response,
                            )
                        ],
                    )
                    results.append(content)

            except Exception as e:
                if return_errors:
                    function_response = types.FunctionResponse(
                        name=tool_name,
                        response={"error": f"Unexpected error: {str(e)}"},
                        id=call_id,
                    )
                    content = types.Content(
                        role="user",
                        parts=[
                            types.Part.from_function_response(
                                name=function_response.name,
                                response=function_response.response,
                            )
                        ],
                    )
                    results.append(content)

        return results

    async def process_function_calls_as_parts(
        self,
        function_calls: list[types.FunctionCall],
        return_errors: bool = True,
    ) -> list[types.Part]:
        """
        Process Gemini's function calls and return just the Part objects.

        This is useful when you want to construct the Content object yourself
        or combine multiple function responses into a single Content.

        Args:
            function_calls: List of FunctionCall objects from the Gemini response.
            return_errors: If True, include error responses for failed calls.

        Returns:
            List of Part objects containing function responses.

        Example:
            parts = await adapter.process_function_calls_as_parts(function_calls)

            # Combine into single Content:
            contents.append(types.Content(role="user", parts=parts))
        """
        results: list[types.Part] = []

        for fc in function_calls:
            tool_name = fc.name
            arguments = fc.args if fc.args else {}

            try:
                result = await self.make_request(
                    f"/tools/{tool_name}/call",
                    method="POST",
                    data={"arguments": arguments},
                )

                part = types.Part.from_function_response(
                    name=tool_name,
                    response={"result": result}
                    if not isinstance(result, dict)
                    else result,
                )
                results.append(part)

            except httpx.HTTPError as e:
                if return_errors:
                    part = types.Part.from_function_response(
                        name=tool_name,
                        response={"error": f"HTTP error calling tool: {str(e)}"},
                    )
                    results.append(part)

            except Exception as e:
                if return_errors:
                    part = types.Part.from_function_response(
                        name=tool_name,
                        response={"error": f"Unexpected error: {str(e)}"},
                    )
                    results.append(part)

        return results

    @staticmethod
    def create_function_response_content(
        name: str,
        response: dict,
    ) -> types.Content:
        """
        Helper to create a function response Content object.

        Args:
            name: The function name that was called.
            response: The response data from the function.

        Returns:
            A Content object that can be appended to the conversation.
        """
        return types.Content(
            role="user",
            parts=[types.Part.from_function_response(name=name, response=response)],
        )



import httpx
import pytest
import respx

from mcphero.adapters.base_adapter import BaseAdapter


class TestBaseAdapterInit:
    def test_stores_base_url(self, base_url):
        adapter = BaseAdapter(base_url)
        assert adapter.base_url == base_url

    def test_strips_trailing_slash(self):
        adapter = BaseAdapter("https://example.com/")
        assert adapter.base_url == "https://example.com"

    def test_default_timeout(self, base_url):
        adapter = BaseAdapter(base_url)
        assert adapter.timeout == 30.0

    def test_custom_timeout(self, base_url):
        adapter = BaseAdapter(base_url, timeout=60.0)
        assert adapter.timeout == 60.0

    def test_default_headers(self, base_url):
        adapter = BaseAdapter(base_url)
        assert adapter.headers == {}

    def test_custom_headers(self, base_url):
        headers = {"Authorization": "Bearer token"}
        adapter = BaseAdapter(base_url, headers=headers)
        assert adapter.headers == headers


class TestMakeRequest:
    @respx.mock
    async def test_get_returns_parsed_json(self, base_url):
        expected = [{"name": "tool1"}]
        respx.get(f"{base_url}/tools").mock(
            return_value=httpx.Response(200, json=expected)
        )

        adapter = BaseAdapter(base_url)
        result = await adapter.make_request("/tools")
        assert result == expected

    @respx.mock
    async def test_post_sends_json_body(self, base_url):
        expected = {"result": "ok"}
        route = respx.post(f"{base_url}/tools/test/call").mock(
            return_value=httpx.Response(200, json=expected)
        )

        adapter = BaseAdapter(base_url)
        result = await adapter.make_request(
            "/tools/test/call", method="POST", data={"arguments": {"q": "hello"}}
        )
        assert result == expected
        assert route.called

    @respx.mock
    async def test_unsupported_method_raises_value_error(self, base_url):
        adapter = BaseAdapter(base_url)
        with pytest.raises(ValueError, match="Unsupported HTTP method: DELETE"):
            await adapter.make_request("/tools", method="DELETE")

    @respx.mock
    async def test_http_error_raises(self, base_url):
        respx.get(f"{base_url}/tools").mock(
            return_value=httpx.Response(500, text="Internal Server Error")
        )

        adapter = BaseAdapter(base_url)
        with pytest.raises(httpx.HTTPStatusError):
            await adapter.make_request("/tools")

    @respx.mock
    async def test_get_passes_params(self, base_url):
        route = respx.get(f"{base_url}/search", params={"q": "test"}).mock(
            return_value=httpx.Response(200, json={"results": []})
        )

        adapter = BaseAdapter(base_url)
        result = await adapter.make_request("/search", params={"q": "test"})
        assert result == {"results": []}
        assert route.called



import pytest


class TestPackageInit:
    def test_openai_adapter_importable(self):
        from mcphero import MCPToolAdapterOpenAI

        assert MCPToolAdapterOpenAI is not None

    def test_all_contains_both_adapters(self):
        import mcphero

        assert "MCPToolAdapterOpenAI" in mcphero.__all__
        assert "MCPToolAdapterGemini" in mcphero.__all__

    def test_invalid_attribute_raises(self):
        import mcphero

        with pytest.raises(AttributeError, match="has no attribute"):
            _ = mcphero.NoSuchAdapter



import httpx
import pytest
import respx

try:
    from google.genai import types

    HAS_GOOGLE_GENAI = True
except ImportError:
    HAS_GOOGLE_GENAI = False

pytestmark = pytest.mark.skipif(
    not HAS_GOOGLE_GENAI, reason="google-genai not installed"
)


@pytest.fixture
def adapter(base_url):
    from mcphero.adapters.gemini import MCPToolAdapterGemini

    return MCPToolAdapterGemini(base_url)


class TestGetFunctionDeclarations:
    @respx.mock
    async def test_converts_mcp_tools(self, base_url, adapter, sample_mcp_tools):
        respx.get(f"{base_url}/tools").mock(
            return_value=httpx.Response(200, json=sample_mcp_tools)
        )

        declarations = await adapter.get_function_declarations()

        assert len(declarations) == 2
        assert declarations[0].name == "get_weather"
        assert declarations[0].description == "Get the current weather for a location"

    @respx.mock
    async def test_handles_missing_schema(
        self, base_url, adapter, sample_mcp_tool_no_schema
    ):
        respx.get(f"{base_url}/tools").mock(
            return_value=httpx.Response(200, json=sample_mcp_tool_no_schema)
        )

        declarations = await adapter.get_function_declarations()

        assert len(declarations) == 1
        assert declarations[0].name == "ping"

    @respx.mock
    async def test_empty_list(self, base_url, adapter):
        respx.get(f"{base_url}/tools").mock(
            return_value=httpx.Response(200, json=[])
        )

        declarations = await adapter.get_function_declarations()
        assert declarations == []


class TestGetTool:
    @respx.mock
    async def test_returns_tool_wrapping_declarations(
        self, base_url, adapter, sample_mcp_tools
    ):
        respx.get(f"{base_url}/tools").mock(
            return_value=httpx.Response(200, json=sample_mcp_tools)
        )

        tool = await adapter.get_tool()

        assert isinstance(tool, types.Tool)
        assert len(tool.function_declarations) == 2


class TestProcessFunctionCalls:
    @respx.mock
    async def test_success(self, base_url, adapter, sample_tool_result):
        respx.post(f"{base_url}/tools/get_weather/call").mock(
            return_value=httpx.Response(200, json=sample_tool_result)
        )

        fc = types.FunctionCall(name="get_weather", args={"location": "London"})
        results = await adapter.process_function_calls([fc])

        assert len(results) == 1
        assert isinstance(results[0], types.Content)
        assert results[0].role == "user"
        assert len(results[0].parts) == 1

    @respx.mock
    async def test_non_dict_result_wrapped(self, base_url, adapter):
        # When result is not a dict, it gets wrapped as {"result": ...}
        respx.post(f"{base_url}/tools/get_weather/call").mock(
            return_value=httpx.Response(200, json="string result")
        )

        fc = types.FunctionCall(name="get_weather", args={"location": "London"})
        results = await adapter.process_function_calls([fc])

        assert len(results) == 1

    @respx.mock
    async def test_dict_result_passed_directly(self, base_url, adapter):
        data = {"temp": 72}
        respx.post(f"{base_url}/tools/get_weather/call").mock(
            return_value=httpx.Response(200, json=data)
        )

        fc = types.FunctionCall(name="get_weather", args={"location": "London"})
        results = await adapter.process_function_calls([fc])

        assert len(results) == 1

    @respx.mock
    async def test_http_error(self, base_url, adapter):
        respx.post(f"{base_url}/tools/get_weather/call").mock(
            return_value=httpx.Response(500, text="Server Error")
        )

        fc = types.FunctionCall(name="get_weather", args={"location": "London"})
        results = await adapter.process_function_calls([fc])

        assert len(results) == 1

    @respx.mock
    async def test_return_errors_false(self, base_url, adapter):
        respx.post(f"{base_url}/tools/get_weather/call").mock(
            return_value=httpx.Response(500, text="Server Error")
        )

        fc = types.FunctionCall(name="get_weather", args={"location": "London"})
        results = await adapter.process_function_calls([fc], return_errors=False)

        assert len(results) == 0

    @respx.mock
    async def test_empty_args(self, base_url, adapter):
        respx.post(f"{base_url}/tools/ping/call").mock(
            return_value=httpx.Response(200, json={"status": "ok"})
        )

        fc = types.FunctionCall(name="ping", args=None)
        results = await adapter.process_function_calls([fc])

        assert len(results) == 1


class TestProcessFunctionCallsAsParts:
    @respx.mock
    async def test_returns_parts(self, base_url, adapter, sample_tool_result):
        respx.post(f"{base_url}/tools/get_weather/call").mock(
            return_value=httpx.Response(200, json=sample_tool_result)
        )

        fc = types.FunctionCall(name="get_weather", args={"location": "London"})
        parts = await adapter.process_function_calls_as_parts([fc])

        assert len(parts) == 1
        assert isinstance(parts[0], types.Part)

    @respx.mock
    async def test_error_returns_part(self, base_url, adapter):
        respx.post(f"{base_url}/tools/get_weather/call").mock(
            return_value=httpx.Response(500, text="Server Error")
        )

        fc = types.FunctionCall(name="get_weather", args={"location": "London"})
        parts = await adapter.process_function_calls_as_parts([fc])

        assert len(parts) == 1
        assert isinstance(parts[0], types.Part)


class TestCreateFunctionResponseContent:
    def test_returns_correct_content(self):
        from mcphero.adapters.gemini import MCPToolAdapterGemini

        content = MCPToolAdapterGemini.create_function_response_content(
            name="get_weather",
            response={"temp": 72},
        )

        assert isinstance(content, types.Content)
        assert content.role == "user"
        assert len(content.parts) == 1



"""
MCP Tool Adapter for OpenAI.

This adapter converts remote MCP server tools to OpenAI-compatible tool definitions
and processes OpenAI's tool calls to HTTP requests to the MCP server.
"""

from __future__ import annotations

import json

import httpx
from openai.types.chat import (
    ChatCompletionMessageToolCall,
    ChatCompletionToolMessageParam,
    ChatCompletionToolParam,
)

from mcphero.adapters.base_adapter import BaseAdapter


class MCPToolAdapterOpenAI(BaseAdapter):
    """
    Adapter that converts remote MCP server tools to OpenAI-compatible tool definitions.
    Also converts OpenAI's tool_calls to HTTP requests to the MCP server.

    Usage:

        from openai import OpenAI
        from mcphero import MCPToolAdapterOpenAI

        adapter = MCPToolAdapterOpenAI("https://api.mcphero.app/mcp/your-server-id")
        client = OpenAI()

        # Get tool definitions
        tools = await adapter.get_tool_definitions()

        # Make request with tools
        response = client.chat.completions.create(
            model="gpt-4o",
            messages=[{"role": "user", "content": "What's the weather?"}],
            tools=tools,
        )

        # Process tool calls if present
        if response.choices[0].message.tool_calls:
            tool_results = await adapter.process_tool_calls(
                response.choices[0].message.tool_calls
            )

            # Continue conversation with results
            messages = [
                {"role": "user", "content": "What's the weather?"},
                response.choices[0].message,
                *tool_results,
            ]
            final_response = client.chat.completions.create(
                model="gpt-4o",
                messages=messages,
                tools=tools,
            )
    """

    async def get_tool_definitions(self) -> list[ChatCompletionToolParam]:
        """
        Fetch tools from MCP server and convert them to OpenAI tool schemas.

        Returns:
            List of tool definitions compatible with OpenAI's `tools` parameter.

        Example:
            tools = await adapter.get_tool_definitions()

            response = client.chat.completions.create(
                model="gpt-4o",
                messages=messages,
                tools=tools,
            )
        """
        mcp_tools = await self.make_request("/tools")

        openai_tools: list[ChatCompletionToolParam] = []
        for tool in mcp_tools:
            openai_tools.append(
                {
                    "type": "function",
                    "function": {
                        "name": tool["name"],
                        "description": tool.get("description", ""),
                        "parameters": tool.get(
                            "inputSchema",
                            {
                                "type": "object",
                                "properties": {},
                            },
                        ),
                    },
                }
            )

        return openai_tools

    async def process_tool_calls(
        self,
        tool_calls: list[ChatCompletionMessageToolCall],
        return_errors: bool = True,
    ) -> list[ChatCompletionToolMessageParam]:
        """
        Process OpenAI's tool_calls by invoking the tools via HTTP.

        Args:
            tool_calls: List of tool calls from `response.choices[0].message.tool_calls`.
            return_errors: If True, include error messages for failed calls.
                If False, failed calls are omitted from results.

        Returns:
            List of tool message dicts compatible with OpenAI's messages format.
            Each result has `role="tool"`, `tool_call_id`, and `content`.

        Example:
            response = client.chat.completions.create(
                model="gpt-4o",
                messages=messages,
                tools=tools,
            )

            if response.choices[0].message.tool_calls:
                tool_results = await adapter.process_tool_calls(
                    response.choices[0].message.tool_calls
                )

                # Add to conversation
                messages.append(response.choices[0].message)
                messages.extend(tool_results)
        """
        results: list[ChatCompletionToolMessageParam] = []

        for tool_call in tool_calls:
            tool_name = tool_call.function.name

            try:
                arguments = json.loads(tool_call.function.arguments)
            except json.JSONDecodeError:
                if return_errors:
                    results.append(
                        {
                            "tool_call_id": tool_call.id,
                            "role": "tool",
                            "content": json.dumps(
                                {"error": "Failed to parse tool arguments"}
                            ),
                        }
                    )
                continue

            try:
                result = await self.make_request(
                    f"/tools/{tool_name}/call",
                    method="POST",
                    data={"arguments": arguments},
                )

                results.append(
                    {
                        "tool_call_id": tool_call.id,
                        "role": "tool",
                        "content": json.dumps(result)
                        if not isinstance(result, str)
                        else result,
                    }
                )

            except httpx.HTTPError as e:
                if return_errors:
                    results.append(
                        {
                            "tool_call_id": tool_call.id,
                            "role": "tool",
                            "content": json.dumps(
                                {"error": f"HTTP error calling tool: {str(e)}"}
                            ),
                        }
                    )

            except Exception as e:
                if return_errors:
                    results.append(
                        {
                            "tool_call_id": tool_call.id,
                            "role": "tool",
                            "content": json.dumps(
                                {"error": f"Unexpected error: {str(e)}"}
                            ),
                        }
                    )

        return results



from __future__ import annotations

import httpx


class BaseAdapter:
    def __init__(
        self,
        base_url: str,
        *,
        timeout: float = 30.0,
        headers: dict | None = None,
    ):
        self.base_url = base_url.rstrip("/")
        self.timeout = timeout
        self.headers = headers or {}

    async def make_request(
        self,
        endpoint: str,
        method: str = "GET",
        data: dict | None = None,
        params: dict | None = None,
    ) -> dict:
        async with httpx.AsyncClient(
            base_url=self.base_url,
            timeout=self.timeout,
            headers=self.headers,
        ) as client:
            if method.upper() == "GET":
                response = await client.get(endpoint, params=params)
            elif method.upper() == "POST":
                response = await client.post(endpoint, json=data)
            else:
                raise ValueError(f"Unsupported HTTP method: {method}")

            response.raise_for_status()
            return response.json()
